﻿--
-- Bunker Silo Worker
--
-- @author  LordBanana
-- @date    2023-07-28
--

BunkerSiloWorker = {};

local specTypeName = 'bunkerSiloWorker'
local modSpecTypeName = g_currentModName ..".".. specTypeName

-- Defines
BunkerSiloWorker.STATE_NONE             = 0
BunkerSiloWorker.STATE_RUNNING			= 1
BunkerSiloWorker.STATE_WAITING          = 2

BunkerSiloWorker.REASON_NONE            = 0
BunkerSiloWorker.REASON_UNSUPPORTED_EQUIPMENT  = 1
BunkerSiloWorker.REASON_UNSUPPORTED_BUNKERSILO  = 2
BunkerSiloWorker.REASON_CLEAR_WARNING   = 3

BunkerSiloWorker.SIDESHIFT_DIRECTION_LEFT    = 1
BunkerSiloWorker.SIDESHIFT_DIRECTION_RIGHT	 = -1
BunkerSiloWorker.SIDESHIFT_STEP_SIZE		 = 0.75

BunkerSiloWorker.SAFETYMARGIN_FRONT_REAR	 = 1.0
BunkerSiloWorker.SAFETYMARGIN_SIDES		 = 1.0

BunkerSiloWorker.MIN_TRAVEL_DISTANCE		 = 8

-- Functions
function BunkerSiloWorker.getSpec(self)
    return self["spec_" .. modSpecTypeName]
end

function BunkerSiloWorker.prerequisitesPresent(specializations)
    return  true  == SpecializationUtil.hasSpecialization(Drivable      ,specializations)
        and true  == SpecializationUtil.hasSpecialization(Motorized     ,specializations)
        and true  == SpecializationUtil.hasSpecialization(Enterable     ,specializations)
        and true  == SpecializationUtil.hasSpecialization(AIVehicle     ,specializations)
        and true  == SpecializationUtil.hasSpecialization(Lights        ,specializations)
        and false == SpecializationUtil.hasSpecialization(ConveyorBelt  ,specializations)
        and false == SpecializationUtil.hasSpecialization(Locomotive    ,specializations)
end

function BunkerSiloWorker.registerFunctions(vehicleType)
    for _,funcName in pairs( {
        "getIsBunkerSiloWorkerActive",
        "canDriveWithAttachedTool",
        "calculateBunkerParams",
        "showReason",
        "setWarning",
      } ) do
        SpecializationUtil.registerFunction(vehicleType, funcName, BunkerSiloWorker[funcName])
    end

    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getAINeedsTrafficCollisionBox", BunkerSiloWorker.getAINeedsTrafficCollisionBox)
end

function BunkerSiloWorker.registerEventListeners(vehicleType)
    for _,funcName in pairs( {
        "onDraw",
        "onLoad",
        "onUpdateTick",
        "onRegisterActionEvents",
        "onAIJobStarted",
        "onAIJobFinished"
       } ) do
        SpecializationUtil.registerEventListener(vehicleType, funcName, BunkerSiloWorker)
    end
end

function BunkerSiloWorker:onLoad(savegame)
    self.bunkerSiloWorkerIsStarted = false

    local spec = BunkerSiloWorker.getSpec(self)

    spec.ShowWarningText = nil;
    spec.ShowWarningTime = 0;
	
	spec.runState = BunkerSiloWorker.STATE_NONE;
	
	spec.origDriveDirection = {0, 0};
	spec.currDriveDirection = {0, 0};
	spec.origDriveTarget = {0, 0};
	spec.currDriveTarget = {0, 0};
	spec.currSideShiftDirection = BunkerSiloWorker.SIDESHIFT_DIRECTION_LEFT;
	spec.lookAheadDistanceFront = 3;
	spec.lookAheadDistanceRear = 3;
	
	spec.currTravelDistance = 0;	-- Current Travel Distance since last reversed direction
	spec.currDistanceToBunkerEdge = 0; -- Current distance to bunker edge
	
	-- Variables to keep track of the vehicle physical limits
	spec.frontBorder = 0;
	spec.rearBorder = 0;
	spec.leftBorder = 0;
	spec.rightBorder = 0;

    -- Variables to keep track of how near we should go to the edges
    spec.safetyMarginFront = BunkerSiloWorker.SAFETYMARGIN_FRONT_REAR;
    spec.safetyMarginRear = BunkerSiloWorker.SAFETYMARGIN_FRONT_REAR;
    spec.safetyMarginSides  = BunkerSiloWorker.SAFETYMARGIN_SIDES;
    spec.showMargins = false;
	
	-- Variables to keep track of where to check for filltype
	spec.ftsX = 0;
	spec.ftsZ = 0;
	spec.ftwX = 0;
	spec.ftwZ = 0;
	spec.fthX = 0;
	spec.fthZ = 0;
	spec.areaChecked = 0;
	spec.fillLevel = 0;
	
	-- Variables to keep track of where to check for filltype on the sides
	spec.sideArea = {};
	spec.sideArea.sX = 0;
	spec.sideArea.sZ = 0;
	spec.sideArea.wX = 0;
	spec.sideArea.wZ = 0;
	spec.sideArea.hX = 0;
	spec.sideArea.hZ = 0;
	spec.sideArea.areaChecked = 0;
	spec.sideArea.fillLevel = 0;
	
	-- Bunker Silo Area parameters, calculated when worker is activated
	spec.bunkerArea = {};
	spec.bunkerArea.startNode = nil;		-- The start nodeID
	spec.bunkerArea.widthNode = nil;		-- The width nodeID
	spec.bunkerArea.heightNode = nil;		-- The height nodeID
    spec.bunkerArea.AINode = nil;           -- The AI nodeID
	spec.bunkerArea.width  = 0;             -- The total width of the bunker
	spec.bunkerArea.height = 0;             -- The total height (length) of the bunker

	spec.turnaroundpoint = {};
	spec.turnaroundpoint.x = 0;
	spec.turnaroundpoint.y = 0;
	spec.turnaroundpoint.z = 0;
	
	spec.debugInfo = {};
	spec.debugInfo.frontDistance = 0;
	spec.debugInfo.rearDistance = 0;
	spec.debugInfo.leftDistance = 0;
	spec.debugInfo.rightDistance = 0;
	spec.debugInfo.vehicleX = 0;
	spec.debugInfo.vehicleZ = 0;
	spec.debugInfo.vX = 0;
	spec.debugInfo.vY = 0;
	spec.debugInfo.vZ = 0;
	spec.debugInfo.tX = 0;
	spec.debugInfo.tY = 0;
	spec.debugInfo.tZ = 0;
	spec.debugInfo.pX = 0;
	spec.debugInfo.pZ = 0;
	
	spec.prevCompactedPercent = 0;
end;

function BunkerSiloWorker:getIsBunkerSiloWorkerActive()
    local spec = BunkerSiloWorker.getSpec(self);
    return spec.runState ~= BunkerSiloWorker.STATE_NONE
end

function BunkerSiloWorker:getAINeedsTrafficCollisionBox(superFunc)
    local spec = BunkerSiloWorker.getSpec(self);
    if spec ~= nil then
        -- We can't use getIsBunkerSiloWorkerActive here as that is
        -- disabled too soon when de-activating the worker and we
        -- get an error message.
        if spec.bunkerSilo ~= nil then
            return false;
        end
    end
    return superFunc(self)
end

function BunkerSiloWorker:canDriveWithAttachedTool()
    -- Locate supported equipment
    if nil ~= self.getAttachedImplements then
		local tools = self:getAttachedImplements();
		if #tools > 0 then
			Debug.log("Unsupported tools attached!");
			return false;
		end
	end
	return true;
end

function BunkerSiloWorker:onUpdateTick(dt, isActiveForInput, isSelected)
    local spec = BunkerSiloWorker.getSpec(self)
	if spec.bunkerSilo ~= nil and BunkerSiloWorker.getIsBunkerSiloWorkerActive(self) then
		-- Check if compaction is complete
        if spec.prevCompactedPercent < 100 and spec.bunkerSilo.compactedPercent >= 100 then
			-- Show notification
			if self.isClient then
				if g_currentMission.player.farmId == self.spec_aiJobVehicle.startedFarmId then
					if self:getCurrentHelper() ~= nil then
						local helperName = self:getCurrentHelper().name
						local text = string.format(g_i18n:getText("BunkerSiloWorker100Complete"), helperName)
						g_currentMission:addIngameNotification(FSBaseMission.INGAME_NOTIFICATION_INFO, text)
					end
				end
			end
		end
		spec.prevCompactedPercent = spec.bunkerSilo.compactedPercent;
	end;
end;

function BunkerSiloWorker:onRegisterActionEvents(isActiveForInput, isActiveForInputIgnoreSelection)
    if self.isClient then
        local spec = BunkerSiloWorker.getSpec(self)
        if spec ~= nil then
            self:clearActionEventsTable(spec.actionEvents)

            if self:getIsEntered() and BunkerSiloWorker.getIsBunkerSiloWorkerActive(self) then
                local _, actionEventId = self:addActionEvent(spec.actionEvents, InputAction.BSW_CHANGE_SIDES_MARGIN, self, BunkerSiloWorker.actionEventChangeSidesDistance, false, true, false, true, nil, nil, true, true)
                g_inputBinding:setActionEventText(actionEventId, g_i18n:getText("input_BSW_CHANGE_SIDES_MARGIN") .. string.format(": %01.1f", spec.safetyMarginSides));
                g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_NORMAL);
                spec.changeSidesDistanceEventId = actionEventId;

                local _, actionEventId3 = self:addActionEvent(spec.actionEvents, InputAction.BSW_CHANGE_FRONT_MARGIN, self, BunkerSiloWorker.actionEventChangeFrontDistance, false, true, false, true, nil, nil, true, true)
                g_inputBinding:setActionEventText(actionEventId3, g_i18n:getText("input_BSW_CHANGE_FRONT_MARGIN") .. string.format(": %01.1f", spec.safetyMarginFront));
                g_inputBinding:setActionEventTextPriority(actionEventId3, GS_PRIO_NORMAL);
                spec.changeFrontDistanceEventId = actionEventId3;

                local _, actionEventId2 = self:addActionEvent(spec.actionEvents, InputAction.BSW_CHANGE_REAR_MARGIN, self, BunkerSiloWorker.actionEventChangeRearDistance, false, true, false, true, nil, nil, true, true)
                g_inputBinding:setActionEventText(actionEventId2, g_i18n:getText("input_BSW_CHANGE_REAR_MARGIN") .. string.format(": %01.1f", spec.safetyMarginRear));
                g_inputBinding:setActionEventTextPriority(actionEventId2, GS_PRIO_NORMAL);
                spec.changeRearDistanceEventId = actionEventId2;

                local _, actionEventId4 = self:addActionEvent(spec.actionEvents, InputAction.BSW_SHOW_HIDE_MARGINS, self, BunkerSiloWorker.actionEventShowHideMargins, false, true, false, true, nil, nil, true, true)
                local text = g_i18n:getText("input_BSW_SHOW_MARGINS");
                if spec.showMargins then
                    text = g_i18n:getText("input_BSW_HIDE_MARGINS");
                end
                g_inputBinding:setActionEventText(actionEventId4, text);
                g_inputBinding:setActionEventTextPriority(actionEventId4, GS_PRIO_NORMAL);
                spec.showHideMarginsEventId = actionEventId4;
            end
        end
    end
end

function BunkerSiloWorker:onAIJobStarted()
    local spec = BunkerSiloWorker.getSpec(self);
    if spec ~= nil then
        if spec.bunkerSilo ~= nil then

            -- We are within a bunker silo, so make this the forced driving strategy
            self.spec_aiVehicle.mod_ForcedDrivingStrategyName = "BunkerSiloWorker"

            -- Check that all implements are valid		
            if BunkerSiloWorker.canDriveWithAttachedTool(self) == false then
                BunkerSiloWorker.showReason(self, connection, BunkerSiloWorker.REASON_UNSUPPORTED_EQUIPMENT)
            elseif BunkerSiloWorker.calculateBunkerParams(self) == false then
                BunkerSiloWorker.showReason(self, connection, BunkerSiloWorker.REASON_UNSUPPORTED_BUNKERSILO)
            else
                BunkerSiloWorker.showReason(self, connection, BunkerSiloWorker.REASON_CLEAR_WARNING)
                spec.runState = BunkerSiloWorker.STATE_RUNNING
                spec.prevCompactedPercent = 0
            end
        end
    end
end;

function BunkerSiloWorker:onAIJobFinished()
     local spec = BunkerSiloWorker.getSpec(self);
     if spec ~= nil then
         spec.runState = BunkerSiloWorker.STATE_NONE;
         self.spec_aiVehicle.mod_ForcedDrivingStrategyName = ""
     end
end

function BunkerSiloWorker.actionEventChangeFrontDistance(self, actionName, inputValue, callbackState, isAnalog)
    local spec = BunkerSiloWorker.getSpec(self);
    if spec ~= nil then
        spec.safetyMarginFront = spec.safetyMarginFront + 0.5;
        if spec.safetyMarginFront > 2.5 then
            spec.safetyMarginFront = 0.5;
        end;
        g_inputBinding:setActionEventText(spec.changeFrontDistanceEventId, g_i18n:getText("input_BSW_CHANGE_FRONT_MARGIN") .. string.format(": %01.1f", spec.safetyMarginFront));
    end;
end;

function BunkerSiloWorker.actionEventChangeRearDistance(self, actionName, inputValue, callbackState, isAnalog)
    local spec = BunkerSiloWorker.getSpec(self);
    if spec ~= nil then
        spec.safetyMarginRear = spec.safetyMarginRear + 0.5;
        if spec.safetyMarginRear > 2.5 then
            spec.safetyMarginRear = 0.5;
        end;
        g_inputBinding:setActionEventText(spec.changeRearDistanceEventId, g_i18n:getText("input_BSW_CHANGE_REAR_MARGIN") .. string.format(": %01.1f", spec.safetyMarginRear));
    end;
end;

function BunkerSiloWorker.actionEventChangeSidesDistance(self, actionName, inputValue, callbackState, isAnalog)
    local spec = BunkerSiloWorker.getSpec(self);
    if spec ~= nil then
        spec.safetyMarginSides = spec.safetyMarginSides + 0.5;
        if spec.safetyMarginSides > 2.5 then
            spec.safetyMarginSides = 0.5;
        end;
        g_inputBinding:setActionEventText(spec.changeSidesDistanceEventId, g_i18n:getText("input_BSW_CHANGE_SIDES_MARGIN") .. string.format(": %01.1f", spec.safetyMarginSides));
    end;
end;

function BunkerSiloWorker.actionEventShowHideMargins(self, actionName, inputValue, callbackState, isAnalog)
    local spec = BunkerSiloWorker.getSpec(self);
    if spec ~= nil then
        spec.showMargins = not spec.showMargins;

        local text = g_i18n:getText("input_BSW_SHOW_MARGINS");
        if spec.showMargins then
            text = g_i18n:getText("input_BSW_HIDE_MARGINS");
        end
        g_inputBinding:setActionEventText(spec.showHideMarginsEventId, text);
    end;
end;

function BunkerSiloWorker:calculateBunkerParams()
	-- Calculate the params needed later

	local spec = BunkerSiloWorker.getSpec(self);
	
	if spec.bunkerSilo == nil then
		-- We're not in any bunker
		Debug.log("Error: Not in any bunker");
		return false;
	end
	
    -- Check that the bunkerSilo actually has a bunkerSiloArea
	if spec.bunkerSilo.bunkerSiloArea == nil then
		-- The bunkerSiloArea is invalid
		Debug.log("Error: BunkerSilo has no valid bunkerSiloArea!");
		return false;
	end
	
    local area = spec.bunkerSilo.bunkerSiloArea;
	
    -- Check that the bunkerSiloArea has the expected nodes
	if area.start == nil or area.height == nil or area.width == nil then
		-- Missing crucial nodes in bunker
		Debug.log("Error: Missing crucial nodes in bunkerSiloArea!");
		return false;
	end

    -- Check that the width/height nodes aren't "sqewed". We expect them
    -- to be in square with the bunker start node coordinate system
    local x, _, z = localToLocal(area.width, area.start, 0, 0, 0);
    if math.abs(x) >= 1 and math.abs(z) >= 1 then
        Debug.log("Error: Width node is not squared with start node coordinate system");
        return false;
    end;
    x, _, z = localToLocal(area.height, area.start, 0, 0, 0);
    if math.abs(x) >= 1 and math.abs(z) >= 1 then
        Debug.log("Error: Height node is not squared with start node coordinate system");
        return false;
    end;
		
	-- Everything seems to be OK!
	spec.bunkerArea.startNode = area.start;
	spec.bunkerArea.heightNode = area.height;
	spec.bunkerArea.widthNode = area.width;
		
	return true;
end

function BunkerSiloWorker:showReason(connection, reason, currentHelper)
    if reason == BunkerSiloWorker.REASON_NONE then
        -- No notification needed
    elseif reason == BunkerSiloWorker.REASON_UNSUPPORTED_EQUIPMENT then
        BunkerSiloWorker.setWarning(self, "BunkerSiloWorkerUnsupportedEquipment");
    elseif reason == BunkerSiloWorker.REASON_UNSUPPORTED_BUNKERSILO then
        BunkerSiloWorker.setWarning(self, "BunkerSiloWorkerUnsupportedBunkerSilo");
    elseif reason == BunkerSiloWorker.REASON_CLEAR_WARNING then
        BunkerSiloWorker.setWarning(self, nil)
    end
end

function BunkerSiloWorker:setWarning(txt, noSendEvent)
    local spec = BunkerSiloWorker.getSpec(self)

    if nil == txt then
        spec.ShowWarningText = "";
        spec.ShowWarningTime = 0;
    else
        spec.ShowWarningText = g_i18n:getText(txt);
        spec.ShowWarningTime = g_currentMission.time + 2500;
    end
end;

function BunkerSiloWorker:drawBunkerMargins()
    local spec = BunkerSiloWorker.getSpec(self)

       -- Draw the silos margin lines
    if spec.bunkerSilo ~= nil then

        -- Find the four corners of the bunker silo + configured margins
        local x1, y1, z1 = localToWorld(spec.bunkerArea.AINode, spec.safetyMarginSides, 0, spec.safetyMarginRear);
        local x2, y2, z2 = localToWorld(spec.bunkerArea.AINode, spec.safetyMarginSides, 0, spec.bunkerArea.height - spec.safetyMarginFront);
        local x3, y3, z3 = localToWorld(spec.bunkerArea.AINode, spec.bunkerArea.width - spec.safetyMarginSides, 0, spec.safetyMarginRear);
        local x4, y4, z4 = localToWorld(spec.bunkerArea.AINode, spec.bunkerArea.width - spec.safetyMarginSides, 0, spec.bunkerArea.height - spec.safetyMarginFront);

        -- Draw lines between these four corners
        drawDebugLine(x1,
                      y1 + 1,
                      z1,
                      1,0,0,
                      x2,
                      y2 + 1,
                      z2,
                      1,0,0, false)
        drawDebugLine(x1,
                      y1 + 1,
                      z1,
                      1,0,0,
                      x3,
                      y3 + 1,
                      z3,
                      1,0,0, false)

        drawDebugLine(x2,
                      y2 + 1,
                      z2,
                      1,0,0,
                      x4,
                      y4 + 1,
                      z4,
                      1,0,0, false)

        drawDebugLine(x3,
                      y3 + 1,
                      z3,
                      1,0,0,
                      x4,
                      y4 + 1,
                      z4,
                      1,0,0, false)
    end
end;

function BunkerSiloWorker:onDraw(isActiveForInput, isSelected)
    local spec = BunkerSiloWorker.getSpec(self)
    if nil == spec then
        return
    end
    if spec.ShowWarningTime > g_currentMission.time then
        g_currentMission:showBlinkingWarning(spec.ShowWarningText)
    end;

    if spec.showMargins and self:getIsEntered() and BunkerSiloWorker.getIsBunkerSiloWorkerActive(self) then
        BunkerSiloWorker.drawBunkerMargins(self);
    end

    if Debug.active then
        setTextColor(1,1,1,1);
        setTextBold(true);
        setTextAlignment(RenderText.ALIGN_LEFT);
        if spec.currDriveDirection ~= nil then
            renderText(0.01, 0.94, 0.015, "currDriveDirection: " .. string.format("%02.2f", spec.currDriveDirection[1]) .. " " .. string.format("%02.2f", spec.currDriveDirection[2]));
        end
        if spec.currDriveTarget ~= nil then
            renderText(0.01, 0.925, 0.015, "currDriveTarget: " .. string.format("%02.2f", spec.currDriveTarget[1]) .. " " .. string.format("%02.2f", spec.currDriveTarget[2]));
        end
        renderText(0.01, 0.91, 0.015, "p: " .. string.format("%02.2f", spec.debugInfo.pX) .. " " .. string.format("%02.2f", spec.debugInfo.pZ));

        renderText(0.01, 0.895, 0.015, "TurningRadius: " .. (self.maxTurningRadius));

        local x = spec.debugInfo.vX;
        local y = spec.debugInfo.vY;
        local z = spec.debugInfo.vZ;	
        local x2 = spec.debugInfo.tX;
        local y2 = spec.debugInfo.tY;
        local z2 = spec.debugInfo.tZ;

        if spec.bunkerSilo ~= nil then
            renderText(0.01, 0.865, 0.015, "FillLevel: " .. spec.fillLevel .. " / " .. spec.areaChecked);
            renderText(0.01, 0.850, 0.015, "SideArea FillLevel: " .. spec.sideArea.fillLevel .. " / " .. spec.sideArea.areaChecked);
            local dx,_,dz = localDirectionToWorld(self:getAIDirectionNode(), 0, 0, 1)
            local vehicleRad = 0.0; MathUtil.getYRotationFromDirection(dx, dz);
            renderText(0.01, 0.835, 0.015, "Rotation: " .. (vehicleRad * 180 / math.pi));
        end;

        renderText(0.01, 0.820, 0.015, "Travel distance: " .. string.format("%02.2f", spec.currTravelDistance));
        renderText(0.01, 0.805, 0.015, "Front distance: " .. string.format("%02.2f", spec.debugInfo.frontDistance));
        renderText(0.01, 0.790, 0.015, "Rear distance: " .. string.format("%02.2f", spec.debugInfo.rearDistance));
        renderText(0.01, 0.775, 0.015, "Left distance: " .. string.format("%02.2f", spec.debugInfo.leftDistance));
        renderText(0.01, 0.760, 0.015, "Right distance: " .. string.format("%02.2f", spec.debugInfo.rightDistance));
        renderText(0.01, 0.745, 0.015, "VehicleXZ: " .. string.format("%02.2f", spec.debugInfo.vehicleX) .. " " .. string.format("%02.2f", spec.debugInfo.vehicleZ));

        drawDebugLine(x,y + 3,z, 1,1,1, x2,y2+3,z2, 0,0,1, false)

        -- Draw the silos debug line
        if spec.bunkerSilo ~= nil then
            drawDebugLine(spec.bunkerSilo.bunkerSiloArea.sx,
                          spec.bunkerSilo.bunkerSiloArea.sy + 5,
                          spec.bunkerSilo.bunkerSiloArea.sz,
                          1,1,1,
                          spec.bunkerSilo.bunkerSiloArea.hx,
                          spec.bunkerSilo.bunkerSiloArea.hy + 5,
                          spec.bunkerSilo.bunkerSiloArea.hz,
                          0,0,1, false)
            drawDebugLine(spec.bunkerSilo.bunkerSiloArea.sx,
                          spec.bunkerSilo.bunkerSiloArea.sy + 5,
                          spec.bunkerSilo.bunkerSiloArea.sz,
                          1,1,1,
                          spec.bunkerSilo.bunkerSiloArea.wx,
                          spec.bunkerSilo.bunkerSiloArea.wy + 5,
                          spec.bunkerSilo.bunkerSiloArea.wz,
                          1,0,0, false)

            drawDebugLine(spec.bunkerSilo.bunkerSiloArea.wx,
                          spec.bunkerSilo.bunkerSiloArea.wy + 5,
                          spec.bunkerSilo.bunkerSiloArea.wz,
                          1,1,1,
                          spec.bunkerSilo.bunkerSiloArea.hx + spec.bunkerSilo.bunkerSiloArea.dwx,
                          spec.bunkerSilo.bunkerSiloArea.hy + 5 + spec.bunkerSilo.bunkerSiloArea.dwy,
                          spec.bunkerSilo.bunkerSiloArea.hz + spec.bunkerSilo.bunkerSiloArea.dwz,
                          0,1,0, false)

            drawDebugLine(spec.bunkerSilo.bunkerSiloArea.hx,
                          spec.bunkerSilo.bunkerSiloArea.hy + 5,
                          spec.bunkerSilo.bunkerSiloArea.hz,
                          1,1,1,
                          spec.bunkerSilo.bunkerSiloArea.wx + spec.bunkerSilo.bunkerSiloArea.dhx,
                          spec.bunkerSilo.bunkerSiloArea.wy + 5 + spec.bunkerSilo.bunkerSiloArea.dhy,
                          spec.bunkerSilo.bunkerSiloArea.wz + spec.bunkerSilo.bunkerSiloArea.dhz,
                          0,1,0, false)
        end

        local vehicleSpec = BunkerSiloWorker.getSpec(self)

        local x1, y1, z1 = localToWorld(self.rootNode, vehicleSpec.leftBorder, 0, 0);
        local x2, y2, z2 = localToWorld(self.rootNode, vehicleSpec.rightBorder, 0, 0);
        local x3, y3, z3 = localToWorld(self.rootNode, 0, 0, vehicleSpec.frontBorder);
        local x4, y4, z4 = localToWorld(self.rootNode, 0, 0, vehicleSpec.rearBorder);

            drawDebugLine(x1,
                          y1,
                          z1,
                          1,1,1,
                          x2,
                          y2,
                          z2,
                          1,1,1, false)
            drawDebugLine(x3,
                          y3,
                          z3,
                          1,1,1,
                          x4,
                          y4,
                          z4,
                          1,1,1, false)

        local x1, y1, z1 = localToWorld(self.rootNode, 5, 0, spec.frontBorder);
        local x2, y2, z2 = localToWorld(self.rootNode, -5, 0, spec.frontBorder);
        local x3, y3, z3 = localToWorld(self.rootNode, 5, 0, spec.rearBorder);
        local x4, y4, z4 = localToWorld(self.rootNode, -5, 0,  spec.rearBorder);

            drawDebugLine(x1,
                          y1,
                          z1,
                          1,0,0,
                          x2,
                          y2,
                          z2,
                          1,0,0, false)
            drawDebugLine(x3,
                          y3,
                          z3,
                          0,0,1,
                          x4,
                          y4,
                          z4,
                          0,0,1, false)

        x1, y1, z1 = localToWorld(self.rootNode, spec.leftBorder, 0, 5);
        x2, y2, z2 = localToWorld(self.rootNode, spec.leftBorder, 0, -5);
        x3, y3, z3 = localToWorld(self.rootNode, spec.rightBorder, 0, 5);
        x4, y4, z4 = localToWorld(self.rootNode, spec.rightBorder, 0,  -5);

            drawDebugLine(x1,
                          y1,
                          z1,
                          1,0,0,
                          x2,
                          y2,
                          z2,
                          1,0,0, false)
            drawDebugLine(x3,
                          y3,
                          z3,
                          0,0,1,
                          x4,
                          y4,
                          z4,
                          0,0,1, false)

        for i=1,#self.spec_wheels.wheels,1 do
            local wheel = self.spec_wheels.wheels[i];
            x1, y1, z1 = localToWorld(wheel.repr, -wheel.width/2, 0, 0);
            x2, y2, z2 = localToWorld(wheel.repr,  wheel.width/2, 0, 0);
            x3, y3, z3 = localToWorld(wheel.repr, 0, 0, -wheel.radius);
            x4, y4, z4 = localToWorld(wheel.repr, 0, 0,  wheel.radius);

            drawDebugLine(x1,
                          y1,
                          z1,
                          1,1,1,
                          x2,
                          y2,
                          z2,
                          1,1,1, false)
            drawDebugLine(x3,
                          y3,
                          z3,
                          1,1,1,
                          x4,
                          y4,
                          z4,
                          1,1,1, false)

            if wheel.additionalWheels ~= nil then
                for _,additionalWheel in pairs(wheel.additionalWheels) do
                    local refNode = wheel.node
                    local width = 0.5 * additionalWheel.width
                    if wheel.repr ~= wheel.driveNode then
                        refNode = wheel.repr
                    end

                    x1, y1, z1 = localToWorld(additionalWheel.wheelTire, width, 0, 0)
                    x2, y2, z2 = localToWorld(additionalWheel.wheelTire, -width, 0, 0)

                    drawDebugLine(x1,
                                  y1,
                                  z1,
                                  1,1,1,
                                  x2,
                                  y2,
                                  z2,
                                  1,1,1, false)
                end
            end

        end			
            if spec.bunkerSilo ~= nil then
                Debug:drawRectangle(spec.ftsX, spec.ftsZ,
                                  spec.ftwX, spec.ftwZ,
                                  spec.fthX, spec.fthZ,
                                  spec.bunkerSilo.bunkerSiloArea.sy,
                                  1, 0, 0);

                Debug:drawRectangle(spec.sideArea.sX, spec.sideArea.sZ,
                                  spec.sideArea.wX, spec.sideArea.wZ,
                                  spec.sideArea.hX, spec.sideArea.hZ,
                                  spec.bunkerSilo.bunkerSiloArea.sy,
                                  0, 0, 1);

                if spec.bunkerArea.startNode ~= nil then
                    x1, y1, z1 = localToWorld(spec.bunkerArea.startNode, 0, 0, 0);
                    x2, y2, z2 = localToWorld(spec.bunkerArea.widthNode, 0, 0, 0);
                    x3, y3, z3 = localToWorld(spec.bunkerArea.heightNode, 0, 0, 0);

                    drawDebugPoint(x1, y1 + 4, z1, 1, 0, 0, 0.5);
                    drawDebugPoint(x2, y2 + 4, z2, 0, 1, 0, 0.5);
                    drawDebugPoint(x3, y3 + 4, z3, 0, 0, 1, 0.5);
                end;
            end;

        if spec.bunkerArea.AINode ~= nil then
            Debug:drawOrigo(spec.bunkerArea.AINode);
        end;

        setTextAlignment(RenderText.ALIGN_LEFT);
        setTextBold(false);
        setTextColor(1,1,1,1);
    end;
end;

Debug.log(("Script loaded: BunkerSiloWorker.lua - from %s (v%s)"):format(g_currentModName, g_modManager:getModByName(g_currentModName).version));
