--
-- Bunker Silo Worker
--
-- @author  LordBanana
-- @date    2023-07-28
--

AIJobVehicle.getCanStartAIVehicle = Utils.overwrittenFunction(AIJobVehicle.getCanStartAIVehicle,
    function(self, superFunc)
        local spec = BunkerSiloWorker.getSpec(self);
        if nil ~= spec then
            if spec.bunkerSilo ~= nil then
                return true
            end
        end
        return superFunc(self)
    end)

AIVehicle.raiseAIEvent = Utils.overwrittenFunction(AIVehicle.raiseAIEvent,
    function(self, superFunc, aiEvt1, aiEvt2)
        if "BunkerSiloWorker" == self.spec_aiVehicle.mod_ForcedDrivingStrategyName then
            -- Don't raise the `aiEvt2`.
            -- This to avoid any attached implements to unfold/start/whatever, when BunkerSiloWorker is activated via AIVehicle.startAIVehicle()
            SpecializationUtil.raiseEvent(self, aiEvt1)
            return
        end
        -- Forced driving-strategy-id was not 'BunkerSiloWorker', so let the original method do what it need to do.
        superFunc(self, aiEvt1, aiEvt2)
    end)


AIFieldWorker.getCanAIFieldWorkerContinueWork = Utils.overwrittenFunction(AIFieldWorker.getCanAIFieldWorkerContinueWork,
    function(self, superFunc)
        if BunkerSiloWorker.getIsBunkerSiloWorkerActive(self) then
            return true;
        end
        return superFunc(self)
    end)

AIFieldWorker.updateAIFieldWorkerDriveStrategies = Utils.overwrittenFunction(AIFieldWorker.updateAIFieldWorkerDriveStrategies,
    function(self, superFunc)
        if "BunkerSiloWorker" == self.spec_aiVehicle.mod_ForcedDrivingStrategyName then
            local spec = self.spec_aiFieldWorker
            if spec.driveStrategies ~= nil and #spec.driveStrategies > 0 then
                for i=#spec.driveStrategies,1,-1 do
                    spec.driveStrategies[i]:delete()
                    table.remove(spec.driveStrategies, i)
                end
                spec.driveStrategies = {}
            end

            local driveStrategyBunkerSilo = AIDriveStrategyBunkerSilo:new()

            driveStrategyBunkerSilo:setAIVehicle(self)
            table.insert(spec.driveStrategies, driveStrategyBunkerSilo)

            return
        end

        -- Forced driving-strategy-id was not 'BunkerSiloWorker', so let the original method do what it need to do.
        superFunc(self)
    end)

AIImplement.getCanImplementBeUsedForAI = Utils.overwrittenFunction(AIImplement.getCanImplementBeUsedForAI,
    function(self, superFunc)
        local spec = BunkerSiloWorker.getSpec(self);
        if spec ~= nil then
            if spec.bunkerSilo ~= nil then
                Debug.log("In a bunker!")
                return true
            end
        end

        return superFunc(self)
    end)

BunkerSilo.getCanInteract = Utils.overwrittenFunction(BunkerSilo.getCanInteract,
    function(self, superFunc, showInformationOnly)
        local baseReturn = superFunc(self, showInformationOnly);

        -- If the superFunction returned false, still check if we are within a vehicle
        -- that a BunkerSiloWorker is controlling, in that case we still want to show the information.
        if baseReturn == false and showInformationOnly == true and g_currentMission.controlledVehicle ~= nil then
            for vehicle in pairs(self.vehiclesInRange) do
                if vehicle == g_currentMission.controlledVehicle then
                    local spec = BunkerSiloWorker.getSpec(g_currentMission.controlledVehicle);
                    if spec ~= nil then
                        if spec.runState ~= BunkerSiloWorker.STATE_NONE then
                            return true;
                        end
                    end
                end
            end
        end;

        return baseReturn;
    end)

BunkerSilo.interactionTriggerCallback = Utils.appendedFunction(BunkerSilo.interactionTriggerCallback,
    function(self, triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
        local vehicle = g_currentMission.nodeToObject[otherShapeId]

        if vehicle ~= nil then
            -- Get the spec and add this unit (if entered)
            local vehicleSpec = BunkerSiloWorker.getSpec(vehicle);

            -- Make sure to create a list in the BunkerSilo in case there are none.
            if self.vehicleList == nil then
                self.vehicleList = {};
            end;

            -- Add this BunkerSilo to the BunkerSiloWorker spec	of this vehicle
            if vehicleSpec ~= nil then
                if onEnter then
                    vehicleSpec.bunkerSilo = self;
                else
                    vehicleSpec.bunkerSilo = nil;
                end

                -- Refresh the implement data to show correct Helper button status
                if vehicle.spec_aiFieldWorker ~= null then
                    vehicle.spec_aiFieldWorker:updateAIFieldWorkerImplementData()
                end
            end

            -- Add this vehicle/object to the BunkerSilo VehicleList, to be used to
            -- check whether another object enters the bunkersilo.
            if onEnter then
                table.insert(self.vehicleList, vehicle);
            else
                for i=#self.vehicleList,1,-1 do
                    if self.vehicleList[i] == vehicle then
                        table.remove(self.vehicleList, i);
                        break;
                    end;
                end
            end
        end
    end);

Debug.log(("Script loaded: override.lua - from %s (v%s)"):format(g_currentModName, g_modManager:getModByName(g_currentModName).version));